/* tkWWW.c Adds World Wide Web commands to a Tcl interpreter
** ===============
**
** Authors:
** Joseph Wang, Department of Astronomy, University of Texas at Austin
** (joe@astro.as.utexas.edu)
**
** History:
** 25 Jul 1992 Version 0.1 (JW)
**             Initial release of tkWWW
**
** 06 Aug 1992 Version 0.2 (JW)
**             Made procedures Private
**             Added comments to conform with WWW Style Guide
**             Added HtPing
**             Added ability to get source
**             Separated HtGoto into HtLoad and HtParseName
**             History recording is now not automatic
**
** 01 Sep 1992 Version 0.3 (JW)
**             Add argument count check for HtLoad to prevent 
**                segmentation violation
**             Removed HtCd (duplicates cd in standard tcl)
**             Removed HtPing
**             Removed HtHistoryList
**             Consolidated HtNext and HtPrevious into HtHistoryMoveBy
**             Removed unneeded global variables
**
** 18 Oct 1992 Version 0.4 (JW)
**             Added GetReference
**             Moved history handling to frontend
**             Added version retreval
**             Removed HtFind
**             Removed HtHistoryRecord
**             Moved reference_mark to GridText.c
**             Removed remaining globals
**             Changed HtGetText to generate commands
**
** 31 Jan 1993 Version 0.5 (JW)
**             Send PARSE_ALL to HtParse
**
** 20 Feb 1993 Version 0.6 (JW)
**             Added HtUncache command
**             Added HtGetIsIndex command
**             HtGetReference returns list of name and href
**             Added HtGetNumberFromTag command
**
** 14 Mar 1993 Version 0.7 (JW)
**             Switch to libWWW 2.0
**             Add internal links
**             Add HtSetSuffix
**             Add HtAddAnchor
**
** 16 Aug 1993 Version 0.9 (JW)
**             Added HtApp variables
**             Did general rewrite
**
** 15 Aug 1994 Created HtMethodHTTP
**
** Copyright:
** Copyright (C) 1992-1993
** Usenet University -- New Network Academy
** Macvicar Institute for Educational Software Development
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* Include c stuff
** ---------------
*/

#ifdef HAVE_ASSERT_H
#include <assert.h>
#endif
#include <stdio.h>
#include <ctype.h>
#if STDC_HEADERS || HAVE_STRING_H
#include <string.h>
#else /* not STDC_HEADERS and not HAVE_STRING_H */
#include <strings.h>
#define strchr index
#define strrchr rindex
#define memcpy(d, s, n) bcopy ((s), (d), (n))
#define memcmp(s1, s2, n) bcmp((s1), (s2), (n))
#define memset(s, ignore, n) bzero((s), (n))
#endif /* not STDC_HEADERS and not HAVE_STRING_H */

#ifdef STDC_HEADERS
#include <sys/types.h>
#include <stdlib.h>
#endif 

/* Include World Wide Web stuff
** ----------------------------
*/

#include <HTAnchor.h>
#include <HTParse.h>
#include <HTAccess.h>
#include <HTML.h>
#include <HText.h>
#include <HTTCP.h>
#include <HTStream.h>
#include <HTFile.h>
#include <HTFWriter.h>
#include <HTextDef.h>
#include <HTAtom.h>
#include <HTPlain.h>
#include <HTMLGen.h>
#include <HTFormat.h>
#include <HTMIME.h>
#include <HTInit.h>
#include <HTAlert.h>
#include <HTAuth.h>
#include <HTTP.h>
#include <TkWWWCmds.h>

/* Include tcl include header
** -------------------------
*/
#include <tcl.h>

/* Some Global Variables
** ---------------------
*/

PUBLIC char * HTAppName = "phoenix";	/* Application name */
PUBLIC char * HTAppVersion = SERVER_VERSION; 	/* Application version */
PUBLIC char HtErrorBuffer[ERROR_BUFSIZ]; /* @@@@@@@@@ */

EXTERN HTPresentation* default_presentation;
					   
/* Macro to check arguments
** ------------------------
** Assumes interpreter is in interp and number of arguments is in argc
*/

#define HtCheckArgc(min,max,function_name) \
    if ((argc < (min)) || (argc > (max))) { \
       Tcl_AppendResult(interp, (function_name), \
			": Incorrect number of arguments" , NULL); \
       return (TCL_ERROR); \
       }

/* Procedures called by tkWWW functions
** ------------------------------------
*/

PRIVATE int HtLoadCmd(dummy, interp, argc, argv)
     ClientData dummy;
     Tcl_Interp *interp;
     int argc;
     char **argv;
{
  HtCheckArgc(2,2, "HtLoad");
  if (!HTLoadAbsolute(argv[1])) {
    Tcl_SetResult(interp, HtErrorBuffer, TCL_STATIC);
    return (TCL_ERROR);
  }

  Tcl_SetResult(interp, HTMainText->output->data, TCL_STATIC);
  return (TCL_OK);
}

PRIVATE int HtMethodHTTPCmd(dummy, interp, argc, argv)
     ClientData dummy;
     Tcl_Interp *interp;
     int argc;
     char **argv;
{
  HtCheckArgc(2,2, "HtMethodHTTP");
  if (strcmp(argv[1], "GET") != 0 && strcmp(argv[1], "POST") != 0)
  {
    Tcl_AppendResult(interp, "HtMethodHTTP: bad option \"", argv[1],
		     "\".  Must be GET or POST.", NULL);
    return (TCL_ERROR);
  }
  HTMethodHTTP(argv[1]);
  return (TCL_OK);
}

PRIVATE int HtParseNameCmd(dummy, interp, argc, argv)
     ClientData dummy;
     Tcl_Interp *interp;
     int argc;
     char **argv;
{
  char *current_address = NULL;
  
  HtCheckArgc(2,3,"HtParseName");
  if (argc == 3 && argv[2] && *(argv[2])) {
    current_address = argv[2];
  } else {
    current_address = NULL;
    StrAllocCopy(current_address, "file:");

/* The following mess is to get the current working directory in a manner
 * that will work on most platforms 
 */

#ifndef HAVE_GETWD  		/* No getwd() on this machine */
#ifdef HAVE_GETCWD		/* System V variant SIGN CHANGED TBL 921006 !! */

    {
      char wd[1024];			/*!! Arbitrary*/
      extern char * getcwd();
      char * result = getcwd(wd, sizeof(wd)); 
      if (result) {

	StrAllocCat (current_address, wd);
      } else {
	fprintf(stderr,
		"HTBrowse: Can't read working directory (getcwd).\n");
      }
    }  /* end if good getcwd result */
	
#else   /* has NO getcwd */

    if (TRACE) 
      fprintf(stderr,
	      "HTBrowse: This platform does not support getwd() or getcwd()\n");
#endif	/* has no getcwd */

#else   /* has getwd */
    {
      char wd[MAXPATHLEN];
      extern char * getwd();
      char * result = getwd(wd);
      if (result) {
	StrAllocCat(current_address, wd);
      } else {
	fprintf(stderr, "HTBrowse: Can't read working directory.\n");
      }
    }
#endif
		
    StrAllocCat(current_address, "/default.html");
  }

  Tcl_SetResult(interp, HTParse(argv[1], current_address, PARSE_ALL), 
		TCL_DYNAMIC);
  if (argc != 3 || !argv[2] || !*(argv[2])) {
    free(current_address);
  }
  
  return (TCL_OK);
}

PRIVATE int HtUncacheCmd (dummy, interp, argc, argv) 
     ClientData dummy;
     Tcl_Interp *interp;
     int argc;
     char **argv;
{
  HTParentAnchor *parent_anchor;
  HTAnchor *anchor;
  HText *document;
  HtCheckArgc(2, 2, "HtUncache");

  anchor = HTAnchor_findAddress(argv[1]);
  if (anchor) {
    parent_anchor = HTAnchor_parent(anchor);
    document = (HText *)HTAnchor_document(parent_anchor);
    if (document) 
      HText_free(document);
  }
  return (TCL_OK);
}
  
/* HtVersion
** ------------
** returns the version of the tkWWW server
*/

PRIVATE int HtVersionCmd(dummy, interp, argc, argv)
     ClientData dummy;
     Tcl_Interp *interp;
     int argc;
     char **argv;
{
  
  Tcl_SetResult(interp, SERVER_VERSION, TCL_STATIC);
  return (TCL_OK);
}

PUBLIC void HTTkSetOutputFile(format, anchor, fnam)
     HTFormat format;
     HTParentAnchor *anchor;
     const char *fnam;
 {
  char *address;
  HText *text = HText_new(anchor);
  Tcl_DString command[1];
  Tcl_DStringInit(command);
  Tcl_DStringAppendElement(command, "tkW3ConfigDisplayFile");
  Tcl_DStringAppendElement(command, HTAtom_name(format));
  Tcl_DStringAppendElement(command, (char *) fnam);
  HTChunkPuts(text->output, command->string);
  HTChunkPuts(text->output, "\n");
  Tcl_DStringFree(command);

  address = HTAnchor_address((HTAnchor *) anchor);
  Tcl_DStringAppendElement(command, "tkW3HtSetName");
  Tcl_DStringAppendElement(command, address);
  Tcl_DStringAppendElement(command, "");
  HTChunkPuts(text->output, command->string);
  HTChunkTerminate(text->output);
  Tcl_DStringFree(command);
  free(address);
}

PRIVATE HTStream * HTTkDisplay(pres, anchor, sink) 
     HTPresentation *pres;
     HTParentAnchor *anchor;
     HTStream *sink;

{
  char *fnam;
  const char * suffix;
  HTStream* this;
  FILE *fp;
  /* Save the file under a suitably suffixed name */

  suffix = HTFileSuffix(pres->rep);

  fnam = tempnam (NULL, "www");
  if (suffix) strcat(fnam, suffix);
    
  fp = fopen (fnam, "w");
  if (!fp) {
    HTAlert("Can't open temporary file!");
    free(fnam);
    return NULL;
  }

  this = HTFWriter_new(fp);
  HTTkSetOutputFile(pres->rep, anchor, fnam);
  free (fnam);
  return this;
}

PUBLIC BOOL HTTkUseInPlace(format_in, format_out)
     HTFormat format_in;
     HTFormat format_out;
{
  int n = HTList_count(HTPresentations);
  int i;
  HTPresentation *pres;
  for (i=0; i<n; i++) {
    pres = HTList_objectAt(HTPresentations, i);
    if (pres->rep == format_in && pres->rep_out == format_out) 
      return (pres->converter == HTTkDisplay) ? TRUE : FALSE;
  }
  if (default_presentation && 
      default_presentation->converter == HTTkDisplay)
    return TRUE;
  return FALSE;

}

PRIVATE int HtSetSuffixCmd(dummy, interp, argc, argv)
     ClientData dummy;
     Tcl_Interp *interp;
     int argc;
     char **argv;
{
  float priority;
  HtCheckArgc(4, 5, "HtSetSuffix");
  priority  = (argc==4) ? 1.0 : atof(argv[4]);
  HTSetSuffix(argv[1], argv[2], argv[3], priority);
  return (TCL_OK);
}

PRIVATE int HtAuthorizationCmd(dummy, interp, argc, argv)
     ClientData dummy;
     Tcl_Interp *interp;
     int argc;
     char **argv;
{
  HtCheckArgc(4, 4, "HtAuthorization");
  if (strcmp(argv[1], "Basic") != 0)
  {
    Tcl_AppendResult(interp, "HtAuthorization: bad option \"", argv[1],
		     "\".  Must be \"Basic\"", NULL);
    return (TCL_ERROR);
  }
  set_user_and_passwd(argv[2], argv[3]);
  return (TCL_OK);
}

/* Add tkWWW commands to a tcl interpreter init_tkWWW(interp)
** -------------
*/

void init_tkWWW(interp) 
     Tcl_Interp *interp;
{
  Tcl_CreateCommand(interp, "HtAuthorization", HtAuthorizationCmd,
		    (ClientData) NULL, (void (*)()) NULL);
  Tcl_CreateCommand(interp, "HtSetSuffix", HtSetSuffixCmd,
		    (ClientData) NULL, (void (*)()) NULL);
  Tcl_CreateCommand(interp, "HtLoad", HtLoadCmd, 
		    (ClientData) NULL, (void (*)()) NULL);
  Tcl_CreateCommand(interp, "HtMethodHTTP", HtMethodHTTPCmd, 
		    (ClientData) NULL, (void (*)()) NULL);
  Tcl_CreateCommand(interp, "HtParseName", HtParseNameCmd,
		    (ClientData) NULL, (void (*)()) NULL);
  Tcl_CreateCommand(interp, "HtUncache", HtUncacheCmd,
		    (ClientData) NULL, (void (*)()) NULL);
  Tcl_CreateCommand(interp, "HtVersion", HtVersionCmd,
		    (ClientData) NULL, (void (*)()) NULL);
  HTSetConversion("www/mime",  "*",           HTMIMEConvert, 	1.0, 0.0, 0.0);
  HTSetConversion("text/html", "text/x-c",    HTMLToC, 	        0.5, 0.0, 0.0);
  HTSetConversion("text/html", "text/plain",  HTMLToPlain, 	0.5, 0.0, 0.0);
  HTSetConversion("text/html", "www/present", HTMLPresent, 	1.0, 0.0, 0.0);
  HTSetConversion("text/plain", "text/html",  HTPlainToHTML,	1.0, 0.0, 0.0);
  HTSetConversion("*",          "www/present",HTTkDisplay,      0.5, 0.0, 0.0);
}
